"""
CGOL
====
A Conway's Game of Life implementation using numpy and pygame.

Rules of Conway's Game of Life:
1. Any live cell with two or three live neighbors survives.
2. Any dead cell with three live neighbors becomes a live cell.
3. All other live cells die in the next generation. Similarly, all other dead cells stay dead.

Github Repo:
https://github.com/INeido/CGOL
"""
from .world import World
from .utils import *
import pygame
import numpy


class Game:
    """Game
    ====

    Contains the functions needed to run the game.

    :param int rw: Width of the Game.
    :param int rh: Height of the Game.
    :param int cs: Size of a cell in pixel.
    :param int ti: Number of times the game shall update in a second (FPS).
    :param tuple ca: Color for alive cells.
    :param tuple cd: Color for dead cells.
    :param tuple cf: Color to fade dead cells to.
    :param tuple cb: Color for background.
    :param str sf: Path of the in-/output file.
    :param bool ps: Game pauses on a stalemate.
    :param bool po: Game pauses when only (2 period) oscillators remain.
    """

    def __init__(self, rw: int, rh: int, gw: int, gh: int, cs: int, ti: int, se: int, ca: tuple, cd: tuple, cf: tuple, cb: tuple, fr: float, fd: float, sf: str, lo: bool, ps: bool, po: bool):
        self.res_width = rw
        self.res_height = rh
        self.cell_size = cs
        self.tickrate = ti
        self.color_alive = numpy.array(ca)
        self.color_dead = numpy.array(cd)
        self.color_fade = numpy.array(cf)
        self.color_background = numpy.array(cb)
        self.save_file = sf
        self.pause_stalemate = ps
        self.pause_oscillators = po

        self.create_world(gw, gh, se, lo, fr, fd)

    def run(self) -> None:
        self.setup_pygame()
        self.game_loop()

    def setup_pygame(self) -> None:
        """Creates and configures pygame instance.
        """
        pygame.init()
        pygame.display.set_caption("CGOL", "hardware")
        self.dis = pygame.display.set_mode((self.res_width, self.res_height), 0, 8)
        self.clock = pygame.time.Clock()

    def get_borders(self) -> None:
        """Determines the visible edges of the grid based on the current viewport.

        This function updates the following instance variables with the calculated values:
            - vis_north: The y-coordinate of the topmost visible row of cells.
            - vis_south: The y-coordinate of the bottommost visible row of cells.
            - vis_west: The x-coordinate of the leftmost visible column of cells.
            - vis_east: The x-coordinate of the rightmost visible column of cells.
            - vis_width: The width of the visible region, in pixels.
            - vis_height: The height of the visible region, in pixels.
        """
        self.vis_north = max(0, int((-self.offset_y) / self.cell_size))
        self.vis_south = max(0, min(self.world.grid_height, self.world.grid_height - int(self.offset_y / self.cell_size) - self.world.grid_height - int(-self.res_height / self.cell_size) + 1))
        self.vis_west = max(0, int((-self.offset_x) / self.cell_size))
        self.vis_east = max(0, min(self.world.grid_width, self.world.grid_width - int(self.offset_x / self.cell_size) - self.world.grid_width - int(-self.res_width / self.cell_size) + 1))
        self.vis_width = (self.vis_east - self.vis_west) * self.cell_size
        self.vis_height = (self.vis_south - self.vis_north) * self.cell_size

    def draw(self) -> None:
        """Converts the 2D Numpy Array of floats to a 3D Numpy Array of integers.
        More specifically it creates a 3rd dimension with depth 3
        holding the RGB values.

        Before:                     After:

        Shape:      (2, 2)          (2, 2, 3)

        Example:    [[0.5  1.  ]    [[[127  72   0]  [255 144   0]]
                     [0.   0.49]]    [[  0   0   0]  [124  70   0]]]

        The colors in this example used were
        Color alive: [255, 144, 0]
        Color dead:  [  0,   0, 0]
        Color fade:  [  0,   0, 0]

        This 3D Array gets scaled up using numpy be the cell size like so:

        Before:
        [[[127  72   0]  [255 144   0]]
         [[  0   0   0]  [124  70   0]]]

        After:
        [[[127  72   0]  [127  72   0]  [255 144   0]  [255 144   0]]
         [[127  72   0]  [127  72   0]  [255 144   0]  [255 144   0]]
         [[  0   0   0]  [  0   0   0]  [124  70   0]  [124  70   0]]
         [[  0   0   0]  [  0   0   0]  [124  70   0]  [124  70   0]]]

        The cell size in this example is 2 pixels, thus the array got
        scaled by a factor of 2 in both axis.
        """
        # Reset the background color
        self.dis.fill(self.color_background)

        # Get the slice of self.world.grid that is actually visible and has to be rendered.
        colors = self.world.grid[self.vis_west:self.vis_east, self.vis_north:self.vis_south][:, :, numpy.newaxis]

        color_dead_arr = numpy.full((colors.shape[0], colors.shape[1], 3), self.color_dead)
        color_alive_arr = numpy.full((colors.shape[0], colors.shape[1], 3), self.color_alive)

        # Set static colors
        colors = numpy.where(colors == 0, color_dead_arr, colors)
        colors = numpy.where(colors == 1, color_alive_arr, colors)
        # Set fading colors using interpolation
        colors = numpy.where((colors > 0) & (colors < 1), (self.color_fade + (self.color_alive - self.color_fade) * colors), colors)

        # Clip final array
        colors = colors.clip(0, 255).astype(int)

        # Scale the array in both axis
        colors = numpy.repeat(numpy.repeat(colors, self.cell_size, axis=1), self.cell_size, axis=0)

        # Create a surface from the array
        sur = pygame.surfarray.make_surface(colors)

        # If the left or top border is not visible, the offset needs to be adjusted.
        if self.vis_west > 0:
            off_x = 0
        else:
            off_x = self.offset_x
        if self.vis_north > 0:
            off_y = 0
        else:
            off_y = self.offset_y
        # Blit the surface to the display
        self.dis.blit(sur, (off_x, off_y))

        # Update display
        pygame.display.flip()

    def center(self) -> None:
        """Updates offsets so that pygame.surface is centered.
        """
        self.offset_x = (self.res_width / 2) - (self.world.grid_width / 2 * self.cell_size)
        self.offset_y = (self.res_height / 2) - (self.world.grid_height / 2 * self.cell_size)

    def create_world(self, grid_width: int, grid_height: int, seed: int, load: bool, fr, fd) -> None:
        """Creates a new World Object.

        Parameters:
        :param int grid_width: Height of the Grid.
        :param int grid_height: Width of the Grid.
        :param int seed: Seed for the array generation. Default is random (-1).
        :param bool load: Boolean indicating whether the last game should be loaded.
        """
        if load:
            try:
                self.world = World(grid_width, grid_height, seed, fr, fd, self.load_grid())
            except Exception as e:
                print("Couldn't load file.", e)
                shutdown(pygame)
        else:
            self.world = World(grid_width, grid_height, seed, fr, fd)

        # Gets correct offsets to center the grid
        self.center()

        # Create the pygame surface in the correct size
        self.get_borders()

    def calc_generation(self) -> None:
        """Calculates and renders the cells.
        """
        # Iterate generations
        self.world.generations += 1

        # Copy is needed because we are updating 'world' in place and we want to save the last full World
        self.world.backup()

        # Update the state of the world
        self.world.update()

        # Catch if the World stopped developing because of a stalemate
        if self.pause_stalemate and self.world.check_stalemate():
            print("\nGame stopped. Reason: Stalemate.")
            self.game_loop(pause=True)

        # Catch if the World stopped developing because only oscillators remain
        if self.pause_oscillators and self.world.check_oscillators():
            print("\nGame stopped. Reason: Only Oscillators remaining.")
            self.game_loop(pause=True)

    def interpolate(self, point0, point1) -> tuple or int:
        """Interpolates between two points.

        :return: Coordinates of interpolated cells.
        :rtype: tuple or int
        """
        # Calculate distance between previous and current position
        distance = numpy.linalg.norm(numpy.array(point1) - numpy.array(point0))

        # Prevent division by zero
        if distance >= 2:
            # Calculate steps in x and y directions
            steps = numpy.stack(numpy.linspace(point0, point1, int(distance / 2)))

            # Calculate coordinates of interpolated cells
            x = (steps[:, 0]-self.offset_x)//self.cell_size
            y = (steps[:, 1]-self.offset_y)//self.cell_size

            return x.astype(int), y.astype(int)
        else:
            # Calculate coordinates of normal cells
            x = (point1[0]-self.offset_x)//self.cell_size
            y = (point1[1]-self.offset_y)//self.cell_size

            return int(x), int(y)

    def game_loop(self, pause=False) -> None:
        """The main loop that runs the game.
        """
        # Flags
        is_looping = True
        running = not pause
        draw = False
        drag = False
        prev_pos = None

        while True:
            self.clock.tick(self.tickrate)

            # Save mouse position
            curr_pos = pygame.mouse.get_pos()

            # Screen drag
            if drag and prev_pos != None:
                self.offset_x, self.offset_y = numpy.add(numpy.subtract(curr_pos, prev_pos), (oldoffset_x, oldoffset_y))
                self.get_borders()

            # Event loop
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    shutdown(pygame)
                # Key events
                elif event.type == pygame.KEYDOWN:
                    # RETURN pressed: Pause game
                    if event.key == pygame.K_RETURN:
                        running = not running
                    # ESCAPE pressed: Close game
                    if event.key == pygame.K_ESCAPE:
                        shutdown(pygame)
                    # Right Arrow pressed: Forward one generation
                    if event.key == pygame.K_RIGHT:
                        self.calc_generation()
                    # R pressed: Reset game
                    if event.key == pygame.K_r:
                        self.world.populate("seed")
                    # F pressed: Fill with random cells
                    if event.key == pygame.K_f:
                        self.world.populate("random")
                    # A pressed: Fill with alive cells
                    if event.key == pygame.K_a:
                        self.world.populate("alive")
                    # D pressed: Fill with dead cells
                    if event.key == pygame.K_d:
                        self.world.populate("dead")
                    # K pressed: Kill alive cells
                    if event.key == pygame.K_k:
                        self.world.populate("kill")
                    # L pressed: Load last saved game
                    if event.key == pygame.K_l:
                        self.world.load_from_csv(load_grid(self.save_file))
                        self.get_borders()
                    # S pressed: Save current game
                    if event.key == pygame.K_s:
                        save_grid(self.world.grid, self.save_file)
                    # C pressed: Center view
                    if event.key == pygame.K_c:
                        self.center()
                        self.get_borders()
                    # P pressed: Save screenshot
                    if event.key == pygame.K_p:
                        pygame.image.save(self.sur, f"{self.get_save_path() + str(self.world.generations)}.png")
                    # + pressed: Extend grid
                    if event.key == pygame.K_PLUS:
                        self.world.extend()
                        self.get_borders()
                    # - pressed: Reduce grid
                    if event.key == pygame.K_MINUS:
                        self.world.reduce()
                        self.get_borders()

                # Mouse events
                elif event.type == pygame.MOUSEBUTTONDOWN:
                    # Left/Right Click: Draw
                    if event.button == 1 or event.button == 3:
                        oldoffset_x, oldoffset_y = self.offset_x, self.offset_y
                        prev_pos = curr_pos
                        draw = True
                    if event.button == 1:
                        draw_color = 1
                    if event.button == 3:
                        draw_color = 0
                    # Middle Mouse: Drag screen
                    if event.button == 2:
                        oldoffset_x, oldoffset_y = self.offset_x, self.offset_y
                        prev_pos = curr_pos
                        drag = True
                elif event.type == pygame.MOUSEBUTTONUP:
                    # Left/Right Click: Draw
                    if event.button == 1 or event.button == 3:
                        prev_pos = None
                        draw = False
                    # Middle Mouse: Drag screen
                    if event.button == 2:
                        prev_pos = None
                        drag = False

                # Zoom
                elif event.type == pygame.MOUSEWHEEL:
                    if event.y == 1:
                        self.cell_size *= 2
                        self.offset_x = curr_pos[0] + (self.offset_x - curr_pos[0]) * 2
                        self.offset_y = curr_pos[1] + (self.offset_y - curr_pos[1]) * 2
                        self.get_borders()
                    elif event.y == -1 and self.cell_size > 1:
                        self.cell_size /= 2
                        self.offset_x = curr_pos[0] + (self.offset_x - curr_pos[0]) / 2
                        self.offset_y = curr_pos[1] + (self.offset_y - curr_pos[1]) / 2
                        self.get_borders()

            # Interpolate to prevent dotted line
            if draw and prev_pos != None:
                x, y = self.interpolate(prev_pos, curr_pos)
                if isinstance(x, int) and isinstance(y, int):
                    if 0 <= x < self.world.grid_width and 0 <= y < self.world.grid_height:
                        self.world.grid[x, y] = draw_color or (0.5 if self.world.grid[x, y] == 1.0 and not draw_color else self.world.grid[x, y])
                elif min(x) >= 0 and max(x) < self.world.grid_width and min(y) >= 0 and max(y) < self.world.grid_height:
                    for xc, yc in zip(x, y):
                        self.world.grid[xc, yc] = draw_color or (0.5 if self.world.grid[xc, yc] == 1.0 and not draw_color else self.world.grid[xc, yc])
                prev_pos = curr_pos

            # Break out of main loop
            if not is_looping:
                break

            # Draw before we start updating the cells
            self.draw()

            # Skip over generation to pause game
            if not running or draw:
                continue

            # Calculate the next generation
            self.calc_generation()

        shutdown(pygame)



import argparse
import csv
import sys
import os


def get_save_path() -> str:
    """Makes sure the 'cgol/images' folder exists and return its path.

    :return: The home directory of the user appended with '/cgol/images'.capitalize()
    :rtype: string
    """
    path = os.path.expanduser("~") + "/cgol/images/"
    if not os.path.exists(path):
        os.makedirs(path)
    return path


def save_grid(grid, save_file) -> bool:
    """Saves the Grid into a CSV file.
    """
    try:
        with open(save_file, "w", newline="") as csvfile:
            writer = csv.writer(csvfile)

            for row in grid:
                writer.writerow(row)
        print("Last tick saved into:", save_file)
        return True
    except Exception as e:
        print("Couldn't save file.", e)
        return False


def load_grid(save_file) -> list:
    """Loads the Grid from a CSV file.

    :return: Rows of the CSV.
    :rtype: array
    """
    with open(save_file, "r") as csvfile:
        reader = csv.reader(csvfile)
        return [row for row in reader]


def shutdown(pygame) -> None:
    pygame.quit()
    try:
        sys.exit(0)
    except SystemExit:
        os._exit(0)


def parse_cli() -> argparse.Namespace:
    try:
        parser = argparse.ArgumentParser(
            prog="CGOL",
            description="Conway's Game of Life")
        parser.add_argument(
            "-rw",
            dest="rw",
            required=False,
            default=1280,
            type=int,
            help="Width of the Game.")
        parser.add_argument(
            "-rh",
            dest="rh",
            required=False,
            default=720,
            type=int,
            help="Height of the Game.")
        parser.add_argument(
            "-ca",
            dest="ca",
            required=False,
            default=(255, 144, 0),
            type=int,
            help="Color for alive cells. 'R G B'",
            nargs='+')
        parser.add_argument(
            "-cd",
            dest="cd",
            required=False,
            default=(0, 0, 0),
            type=int,
            help="Color for dead cells. 'R G B'",
            nargs='+')
        parser.add_argument(
            "-cf",
            dest="cf",
            required=False,
            default=(0, 0, 0),
            type=int,
            help="Color to fade dead cells to. 'R G B'",
            nargs='+')
        parser.add_argument(
            "-cb",
            dest="cb",
            required=False,
            default=(16, 16, 16),
            type=int,
            help="Color for dead cells. 'R G B'",
            nargs='+')
        parser.add_argument(
            "-cs",
            dest="cs",
            required=False,
            default=8,
            type=int,
            help="Size of a cell in pixel.")
        parser.add_argument(
            "-gw",
            dest="gw",
            required=False,
            default=160,
            type=int,
            help="Width of the World.")
        parser.add_argument(
            "-gh",
            dest="gh",
            required=False,
            default=90,
            type=int,
            help="Height of the World.")
        parser.add_argument(
            "-ti",
            dest="ti",
            required=False,
            default=60,
            type=float,
            help="Number of times the game shall update in a second (FPS).")
        parser.add_argument(
            "-se",
            dest="se",
            required=False,
            default=-1,
            type=int,
            help="Seed value used to create World.")
        parser.add_argument(
            "-sf",
            dest="sf",
            required=False,
            default="./cgol.csv",
            type=str,
            help="Path of the in-/output file. (Should be .csv)")
        parser.add_argument(
            "-lo",
            dest="lo",
            required=False,
            default=False,
            type=bool,
            help="Load revious save.")
        parser.add_argument(
            "-ps",
            dest="ps",
            required=False,
            default=False,
            type=bool,
            help="Game pauses on a stalemate.")
        parser.add_argument(
            "-po",
            dest="po",
            required=False,
            default=False,
            type=bool,
            help="Game pauses when only oscillators remain.")
        parser.add_argument(
            "-fr",
            dest="fr",
            required=False,
            default=0.01,
            type=float,
            help="Value by which a cell should decrease every generation.")
        parser.add_argument(
            "-fd",
            dest="fd",
            required=False,
            default=0.5,
            type=float,
            help="Value a cell should have after death.")
        args = parser.parse_args()
        return args
    except argparse.ArgumentError as err:
        sys.exit(2)


def get_configuration(args: argparse.Namespace) -> dict:
    return {
        "rw": args.rw,
        "rh": args.rh,
        "gw": args.gw,
        "gh": args.gh,
        "cs": args.cs,
        "ti": args.ti,
        "se": args.se,
        "ca": tuple(args.ca),
        "cd": tuple(args.cd),
        "cf": tuple(args.cf),
        "cb": tuple(args.cb),
        "fr": args.fr,
        "fd": args.fd,
        "sf": args.sf,
        "lo": args.lo,
        "ps": args.ps,
        "po": args.po,
    }



"""
World
====
"""
import numpy


class World:
    """World
    ====

    Contains various functions to store and update a grid with Cells.

    :param int gw: Width of the World.
    :param int gh: Height of the World.
    :param int se: Seed for the array generation. Default is random.
    :param float fr: Value a cell should loose per generation after death.
    :param float fd: Fade value a cell should start with after death.
    :param array rows: The 2D Array filled with random 0s and 1s with the last being settings.
    """

    def __init__(self, gw: int, gh: int, se: int, fr: float, fd: float, rows=[]):
        self.grid_width = gw
        self.grid_height = gh
        self.fade_rate = fr
        self.fade_dead = fd
        self.seed = numpy.random.randint(2**32 - 1) if se == -1 else se
        self.generations = 0

        # If 'rows' are empty, create new grid, else convert 'rows' to numpy array.
        self.populate("seed") if len(rows) == 0 else self.load_from_csv(rows)
        self.grid_backup_0 = numpy.zeros_like(self.grid)
        self.grid_backup_1 = numpy.zeros_like(self.grid)

    def populate(self, mode: str) -> None:
        """Fill 'grid' with different values.

        :param bool mode: The mode based on which the array should be filled.
        """
        if mode == "seed":
            self.grid = numpy.random.default_rng(self.seed).choice([0.0, 1.0], size=(self.grid_width, self.grid_height), p=[0.75, 0.25])
        if mode == "random":
            self.grid = numpy.random.choice([0.0, 1.0], size=(self.grid_width, self.grid_height), p=[0.75, 0.25])
        elif mode == "alive":
            self.grid = numpy.ones((self.grid_width, self.grid_height), dtype=float)
        elif mode == "dead":
            self.grid = numpy.zeros((self.grid_width, self.grid_height), dtype=float)
        elif mode == "kill":
            self.grid[self.grid == 1.0] = self.fade_dead
        else:
            return False

    def load_from_csv(self, grid) -> None:
        """Loads data from CSV file.

        :param array grid: The 2D Array filled with random 0s and 1s with the last being settings.
        """
        self.grid = numpy.array(grid, dtype=float)

    def backup(self) -> None:
        """Creates a shallow copy of the grid.
        """
        temp = numpy.copy(self.grid_backup_0)
        self.grid_backup_0 = numpy.copy(self.grid)
        self.grid_backup_1 = numpy.copy(temp)

    def check_stalemate(self) -> bool:
        """Compares the last backup with the current grid to see of it changed.

        :return: Is the backup the same as the current grid?
        :rtype: bool
        """
        return numpy.array_equal(self.grid, self.grid_backup_0)

    def check_oscillators(self) -> bool:
        """Compares the second last backup with the current grid to see of it changed.

        :return: Is the second last backup the same as the current grid?
        :rtype: bool
        """
        return numpy.array_equal(self.grid, self.grid_backup_1)

    def extend(self) -> None:
        """Extends grid in every direction by one row/column.
        """
        self.grid = numpy.pad(self.grid, pad_width=1, mode='constant', constant_values=0)
        self.grid_width += 2
        self.grid_height += 2

    def reduce(self) -> None:
        """Reduce the size of the grid by removing the outermost row and column on all sides.
        If the grid has a width or height less than 3, the function returns without modifying the grid.
        """
        if len(self.grid) < 3 or len(self.grid[0]) < 3:
            return
        self.grid = self.grid[1:-1, 1:-1]
        self.grid_width -= 2
        self.grid_height -= 2

    def get_neighbors(self) -> numpy.array:
        """Gets the number of alive neighbors of a cell in a toroidal space.
        It uses the numpy roll function to shift the values in the
        grid along the x and y axis. 
        The resulting shifted arrays are summed and returned as the
        number of neighbors for each cell.

        The faded values are clipped so that they become integers.

        self.grid:          clipped_grid:
            [[1. 0.2 1. ]       [[1 0 1]
             [0. 1.  0.4]   ->   [0 1 0]
             [1. 0.  1. ]]       [1 0 1]]

        Afterwards the grid is rolled and added onto 'neighbors'.

        neighbors after the rolls:
        1.  [[1 0 0] -> 2.  [[1 0 0] -> 3.  [[1 1 1] ->
             [0 1 1]         [1 1 2]         [2 2 2]
             [0 1 1]]        [1 1 2]]        [2 2 2]]

        4.  [[1 2 2] ->                 5.  [[2 3 2] ->
             [3 2 2]          skip           [3 2 3]
             [2 3 3]]                        [3 4 3]]

        6.  [[2 4 3] -> 7.  [[3 4 4] -> 8.  [[4 5 4]
             [3 3 4]         [4 3 5]         [5 4 5]
             [4 4 3]]        [4 5 3]]        [4 5 4]]

        :return: An array with the count of alive neighbors for each cell.
        :rtype: numpy.array int
        """
        # Create a new array the same size as 'grid'
        neighbors = numpy.zeros_like(self.grid, dtype=int)

        # Convert faded values to zeros
        clipped_grid = numpy.where(self.grid < 1, 0, 1)

        # Roll over every axis to get a new array with the number of neighbors
        for dx in [-1, 0, 1]:
            for dy in [-1, 0, 1]:
                # Dont roll of both directions are 0.
                if not dx and not dy:
                    continue
                neighbors += numpy.roll(numpy.roll(clipped_grid, dx, axis=0), dy, axis=1)

        return neighbors

    def apply_rules_fade(self, neighbors) -> numpy.array:
        """Determines the new state of each cell for the current tick using the "fade" implementation.
        In this implementation, cell values are stored as floats:
            1.0 = alive
            < 1.0 || > 0.0 = fading
            0.0 = dead

        For cells that are currently alive, the rules are applied as follows:
            If the number of neighbors is 2 or 3, the cell remains alive (value = 1.0).
            Otherwise, the cell is considered dead and its value is set to the "fade_dead" value.

        For cells that are currently dead, the rules are applied as follows:
            If the number of neighbors is 3, the cell becomes alive (value = 1.0).
            Otherwise, the cell's value is decreased by the "fade_rate" value.

        :param numpy.array neighbors: The number of neighbors for each cell.
        :return: New state of cells.
        :rtype: numpy.array float
        """
        # Create a copy of the grid to store the next generation
        next_generation = numpy.copy(self.grid)

        # Find the indices of cells that are currently alive
        alive = numpy.where(self.grid == 1)

        # Find the indices of cells that are currently dead
        dead = numpy.where(self.grid < 1)

        # Apply the rules to cells that are currently alive
        next_generation[alive] = numpy.where((neighbors[alive] == 2) | (neighbors[alive] == 3), 1.0, self.fade_dead)

        # Apply the rule to cells that are currently dead
        next_generation[dead] = numpy.where(neighbors[dead] == 3, 1.0, self.grid[dead] - self.fade_rate)

        return numpy.where(next_generation < 0.00001, 0.0, next_generation)

    def apply_rules(self, neighbors) -> numpy.array:
        """Apply Rules
        ====

        Determines the new state of each cell for the current tick.

        :param numpy.array neighbors: The number of neighbors for each cell.

        :return: New state of cells.
        :rtype: numpy.array int
        """
        # Create a copy of the grid to store the next generation
        next_generation = numpy.copy(self.grid)

        # Find the indices of cells that are currently alive
        alive = numpy.where(self.grid == 1)

        # Find the indices of cells that are currently dead
        dead = numpy.where(self.grid == 0)

        # Apply the rules to cells that are currently alive
        next_generation[alive] = numpy.where((neighbors[alive] == 2) | (neighbors[alive] == 3), 1, 0)

        # Apply the rule to cells that are currently dead
        next_generation[dead] = numpy.where(neighbors[dead] == 3, 1, 0)

        return next_generation

    def update(self) -> None:
        """Updates the state of the cells in the world according to the rules of the Game of Life.
        """
        # Get neighbors
        neighbors = self.get_neighbors()

        # Apply rules
        self.grid = self.apply_rules_fade(neighbors)
